//
// SPDX-FileCopyrightText: Copyright 2024 Arm Limited and/or its affiliates <open-source-office@arm.com>
//
// SPDX-License-Identifier: Apache-2.0
//

#if !defined(__aarch64__)
#error This file must be compiled for AArch64.
#else  // Architectural features check.

#include "kai_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla.h"

#include <stddef.h>
#include <stdint.h>

#include "kai/kai_common.h"

static const size_t kai_mr = 6;
static const size_t kai_nr = 8;
static const size_t kai_kr = 1;
static const size_t kai_sr = 1;

size_t kai_get_m_step_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(void) {
    return kai_mr;
}

size_t kai_get_n_step_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(void) {
    return kai_nr;
}

size_t kai_get_nr_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(void) {
    return kai_nr;
}

size_t kai_get_kr_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(void) {
    return kai_kr;
}

size_t kai_get_sr_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(void) {
    return kai_sr;
}

size_t kai_get_lhs_offset_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(size_t m_idx, size_t stride) {
    KAI_ASSUME(m_idx % kai_mr == 0);

    return m_idx * stride;
}

size_t kai_get_rhs_packed_offset_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(size_t n_idx, size_t k) {
    KAI_ASSUME(n_idx % kai_nr == 0);

    return n_idx / kai_nr * (kai_nr * sizeof(float) + kai_nr * k * sizeof(float));
}

size_t kai_get_dst_offset_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(
    size_t m_idx, size_t n_idx, size_t stride) {
    KAI_ASSUME(m_idx % kai_mr == 0);
    KAI_ASSUME(n_idx % kai_nr == 0);

    return m_idx * stride + n_idx * sizeof(float);
}

size_t kai_get_dst_size_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(size_t m, size_t n) {
    return m * n * sizeof(float);
}

void kai_run_matmul_clamp_f32_f32_f32p8x1biasf32_6x8x4_neon_mla(
    size_t m, size_t n, size_t k,                             //
    const void* lhs, size_t lhs_stride,                       //
    const void* rhs_packed,                                   //
    void* dst, size_t dst_stride_row, size_t dst_stride_col,  //
    float clamp_min, float clamp_max) {
    KAI_ASSERT(dst_stride_col == sizeof(float));

    typedef struct {
        float maxval;
        float minval;
        unsigned int num_strings;
        const unsigned int* string_lengths;
        size_t N;
        const void* B_ptr;
        size_t output_offset;
        size_t input_initial_col;
        size_t input_offset;
        void* output_ptr;
        const void* bias;
    } KernelArgs;

    KernelArgs ka;

    unsigned long flags = 0;

    unsigned int string_length = k;
    ka.num_strings = 1;
    ka.string_lengths = &string_length;
    ka.N = n;
    ka.B_ptr = rhs_packed;
    ka.bias = NULL;

    // Direct input.
    const void* input_ptr = lhs;
    ka.input_offset = lhs_stride / sizeof(float);
    ka.input_initial_col = 0;

    // Direct output.
    ka.output_ptr = dst;
    ka.output_offset = dst_stride_row / sizeof(float);

    // Clamping output.
    flags |= 0x2;
    ka.maxval = clamp_max;
    ka.minval = clamp_min;

    __asm__ __volatile__(
        "1:"  // Row loop
        "cmp %x[m], #0x6\n"
        "bge 126f\n"
        "cmp %x[m], #0x4\n"
        "bgt 101f\n"
        "beq 76f\n"
        "cmp %x[m], #0x2\n"
        "bgt 51f\n"
        "beq 26f\n"
        "ldr x11, [%x[args_ptr], %[offsetof_N]]\n"
        "ldr x10, [%x[args_ptr], %[offsetof_B_ptr]]\n"
        "ldr x9, [%x[args_ptr], %[offsetof_output_ptr]]\n"
        "2:"  // Height 1: Column loop
        "cbz x10, 3f\n"
        "ldr q20, [x10, #0x0]\n"
        "ldr q21, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "b 10f\n"
        "3:"  // Height 1: no bias
        "tbz %x[flags], #0, 9f\n"
        "cmp x11, #0x8\n"
        "bge 8f\n"
        "tbz x11, #2, 5f\n"
        "ld1 { v20.4s }, [x9], #0x10\n"
        "tbz x11, #1, 4f\n"
        "ldr d21, [x9], #0x8\n"
        "mov x20, #0x18\n"
        "tbz x11, #0, 7f\n"
        "ld1 { v21.s }[2], [x9]\n"
        "b 7f\n"
        "4:"  // Height 1: Partial accumulate: partial_1_4
        "mov x20, #0x10\n"
        "tbz x11, #0, 7f\n"
        "ldr s21, [x9, #0x0]\n"
        "b 7f\n"
        "5:"  // Height 1: Partial accumulate: partial_2_0
        "tbz x11, #1, 6f\n"
        "ldr d20, [x9], #0x8\n"
        "mov x20, #0x8\n"
        "tbz x11, #0, 7f\n"
        "ld1 { v20.s }[2], [x9]\n"
        "b 7f\n"
        "6:"  // Height 1: Partial accumulate: partial_1_0
        "ldr s20, [x9, #0x0]\n"
        "mov x20, #0x0\n"
        "7:"  // Height 1: Partial accumulate: Done
        "sub x9, x9, x20\n"
        "b 10f\n"
        "8:"  // Height 1: full accumulate
        "ldr q20, [x9, #0x0]\n"
        "ldr q21, [x9, #0x10]\n"
        "b 10f\n"
        "9:"  // Height 1: no accumulate
        "movi v20.16b, #0x0\n"
        "movi v21.16b, #0x0\n"
        "10:"  // Height 1: setup done
        "mov x28, #0x0\n"
        "11:"  // Height 1: String loop
        "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
        "ldr x21, [%x[args_ptr], %[offsetof_input_offset]]\n"
        "ldr w27, [x20, x28, LSL #0x2]\n"
        "tbz %x[flags], #3, 12f\n"
        "ldr x20, [%x[input_ptr], x28, LSL #0x3]\n"
        "add x20, x20, x21, LSL #3\n"
        "ldr x26, [x20, #0x0]\n"
        "cbnz x28, 13f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
        "add x26, x26, x20, LSL #2\n"
        "b 13f\n"
        "12:"  // Height 1: setup direct input
        "mov x26, %x[input_ptr]\n"
        "13:"  // Height 1: input setup done
        "cmp x27, #0x4\n"
        "blt 16f\n"
        "ldr q0, [x26, #0x0]\n"
        "ldr q6, [x10, #0x0]\n"
        "cmp x27, #0x8\n"
        "ldr q7, [x10, #0x10]\n"
        "ldr q8, [x10, #0x20]\n"
        "ldr q9, [x10, #0x30]\n"
        "ldr q10, [x10, #0x40]\n"
        "ldr q11, [x10, #0x50]\n"
        "ldr q12, [x10, #0x60]\n"
        "ldr q13, [x10, #0x70]\n"
        "blt 15f\n"
        "14:"  // Height 1: Multiply loop: Main loop head
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "sub x27, x27, #0x4\n"
        "add x26, x26, #0x10\n"
        "cmp x27, #0x8\n"
        "add x10, x10, #0x80\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "ldr q6, [x10, #0x0]\n"
        "ldr q7, [x10, #0x10]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "ldr q8, [x10, #0x20]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "ldr q9, [x10, #0x30]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "ldr q10, [x10, #0x40]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "ldr q11, [x10, #0x50]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "ldr q12, [x10, #0x60]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "ldr q0, [x26, #0x0]\n"
        "ldr q13, [x10, #0x70]\n"
        "bge 14b\n"
        "15:"  // Height 1: Multiply loop: Single iteration only
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "add x26, x26, #0x10\n"
        "sub x27, x27, #0x4\n"
        "add x10, x10, #0x80\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "16:"  // Height 1: Multiply loop: Main loop skip
        "cbz x27, 18f\n"
        "17:"  // Height 1: Multiply loop: Odd block loop
        "ldr s18, [x26], #0x4\n"
        "ldr q17, [x10, #0x0]\n"
        "sub x27, x27, #0x1\n"
        "ldr q16, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "fmla v20.4s, v17.4s, v18.s[0]\n"
        "fmla v21.4s, v16.4s, v18.s[0]\n"
        "cbnz x27, 17b\n"
        "18:"  // Height 1: Multiply loop: No odd multiplies
        "ldr w20, [%x[args_ptr], %[offsetof_num_strings]]\n"
        "add x28, x28, #0x1\n"
        "cmp x28, x20\n"
        "bne 11b\n"
        "prfm pstl1keep, [x9, #0x0]\n"
        "tbz %x[flags], #1, 19f\n"
        "add x21, %x[args_ptr], %[offset_max]\n"
        "add x20, %x[args_ptr], %[offset_min]\n"
        "ld1r { v17.4s }, [x21]\n"
        "ld1r { v16.4s }, [x20]\n"
        "fmin v20.4s, v20.4s, v17.4s\n"
        "fmin v21.4s, v21.4s, v17.4s\n"
        "fmax v20.4s, v20.4s, v16.4s\n"
        "fmax v21.4s, v21.4s, v16.4s\n"
        "19:"  // Height 1: No activation
        "cmp x11, #0x8\n"
        "bge 24f\n"
        "tbz x11, #2, 21f\n"
        "st1 { v20.4s }, [x9], #0x10\n"
        "tbz x11, #1, 20f\n"
        "str d21, [x9], #0x8\n"
        "tbz x11, #0, 23f\n"
        "st1 { v21.s }[2], [x9]\n"
        "b 23f\n"
        "20:"  // Height 1: Partial direct writeback: partial_1_4
        "tbz x11, #0, 23f\n"
        "str s21, [x9, #0x0]\n"
        "b 23f\n"
        "21:"  // Height 1: Partial direct writeback: partial_2_0
        "tbz x11, #1, 22f\n"
        "str d20, [x9], #0x8\n"
        "tbz x11, #0, 23f\n"
        "st1 { v20.s }[2], [x9]\n"
        "b 23f\n"
        "22:"  // Height 1: Partial direct writeback: partial_1_0
        "str s20, [x9, #0x0]\n"
        "23:"  // Height 1: Partial direct writeback: Done
        "b 25f\n"
        "24:"  // Height 1: Full writeback
        "str q20, [x9, #0x0]\n"
        "str q21, [x9, #0x10]\n"
        "add x9, x9, #0x20\n"
        "25:"  // Height 1: Writeback done
        "subs x11, x11, #0x8\n"
        "bgt 2b\n"
        "b 152f\n"
        "26:"  // Height 2
        "ldr x11, [%x[args_ptr], %[offsetof_N]]\n"
        "ldr x10, [%x[args_ptr], %[offsetof_B_ptr]]\n"
        "ldr x9, [%x[args_ptr], %[offsetof_output_ptr]]\n"
        "27:"  // Height 2: Column loop
        "cbz x10, 28f\n"
        "ldr q20, [x10, #0x0]\n"
        "ldr q21, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "mov v22.16b, v20.16b\n"
        "mov v23.16b, v21.16b\n"
        "b 35f\n"
        "28:"  // Height 2: no bias
        "tbz %x[flags], #0, 34f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "cmp x11, #0x8\n"
        "add x26, x9, x20, LSL #2\n"
        "bge 33f\n"
        "tbz x11, #2, 30f\n"
        "ld1 { v20.4s }, [x9], #0x10\n"
        "ld1 { v22.4s }, [x26], #0x10\n"
        "tbz x11, #1, 29f\n"
        "ldr d21, [x9], #0x8\n"
        "ldr d23, [x26], #0x8\n"
        "mov x20, #0x18\n"
        "tbz x11, #0, 32f\n"
        "ld1 { v21.s }[2], [x9]\n"
        "ld1 { v23.s }[2], [x26]\n"
        "b 32f\n"
        "29:"  // Height 2: Partial accumulate: partial_1_4
        "mov x20, #0x10\n"
        "tbz x11, #0, 32f\n"
        "ldr s21, [x9, #0x0]\n"
        "ldr s23, [x26, #0x0]\n"
        "b 32f\n"
        "30:"  // Height 2: Partial accumulate: partial_2_0
        "tbz x11, #1, 31f\n"
        "ldr d20, [x9], #0x8\n"
        "ldr d22, [x26], #0x8\n"
        "mov x20, #0x8\n"
        "tbz x11, #0, 32f\n"
        "ld1 { v20.s }[2], [x9]\n"
        "ld1 { v22.s }[2], [x26]\n"
        "b 32f\n"
        "31:"  // Height 2: Partial accumulate: partial_1_0
        "ldr s20, [x9, #0x0]\n"
        "ldr s22, [x26, #0x0]\n"
        "mov x20, #0x0\n"
        "32:"  // Height 2: Partial accumulate: Done
        "sub x9, x9, x20\n"
        "b 35f\n"
        "33:"  // Height 2: full accumulate
        "ldr q20, [x9, #0x0]\n"
        "ldr q21, [x9, #0x10]\n"
        "ldr q22, [x26, #0x0]\n"
        "ldr q23, [x26, #0x10]\n"
        "b 35f\n"
        "34:"  // Height 2: no accumulate
        "movi v20.16b, #0x0\n"
        "movi v21.16b, #0x0\n"
        "movi v22.16b, #0x0\n"
        "movi v23.16b, #0x0\n"
        "35:"  // Height 2: setup done
        "mov x28, #0x0\n"
        "36:"  // Height 2: String loop
        "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
        "ldr x21, [%x[args_ptr], %[offsetof_input_offset]]\n"
        "ldr w27, [x20, x28, LSL #0x2]\n"
        "tbz %x[flags], #3, 37f\n"
        "ldr x20, [%x[input_ptr], x28, LSL #0x3]\n"
        "add x20, x20, x21, LSL #3\n"
        "ldr x26, [x20, #0x0]\n"
        "ldr x25, [x20, #0x8]\n"
        "cbnz x28, 38f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
        "add x26, x26, x20, LSL #2\n"
        "add x25, x25, x20, LSL #2\n"
        "b 38f\n"
        "37:"  // Height 2: setup direct input
        "mov x26, %x[input_ptr]\n"
        "add x25, x26, x21, LSL #2\n"
        "38:"  // Height 2: input setup done
        "cmp x27, #0x4\n"
        "blt 41f\n"
        "ldr q0, [x26, #0x0]\n"
        "ldr q1, [x25, #0x0]\n"
        "cmp x27, #0x8\n"
        "ldr q6, [x10, #0x0]\n"
        "ldr q7, [x10, #0x10]\n"
        "ldr q8, [x10, #0x20]\n"
        "ldr q9, [x10, #0x30]\n"
        "ldr q10, [x10, #0x40]\n"
        "ldr q11, [x10, #0x50]\n"
        "ldr q12, [x10, #0x60]\n"
        "ldr q13, [x10, #0x70]\n"
        "blt 40f\n"
        "39:"  // Height 2: Multiply loop: Main loop head
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "sub x27, x27, #0x4\n"
        "add x26, x26, #0x10\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "add x25, x25, #0x10\n"
        "cmp x27, #0x8\n"
        "add x10, x10, #0x80\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "ldr q6, [x10, #0x0]\n"
        "ldr q7, [x10, #0x10]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "ldr q8, [x10, #0x20]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "ldr q9, [x10, #0x30]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "ldr q10, [x10, #0x40]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "ldr q11, [x10, #0x50]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "ldr q12, [x10, #0x60]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "ldr q0, [x26, #0x0]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "ldr q1, [x25, #0x0]\n"
        "ldr q13, [x10, #0x70]\n"
        "bge 39b\n"
        "40:"  // Height 2: Multiply loop: Single iteration only
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "add x26, x26, #0x10\n"
        "add x25, x25, #0x10\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "sub x27, x27, #0x4\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "add x10, x10, #0x80\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "41:"  // Height 2: Multiply loop: Main loop skip
        "cbz x27, 43f\n"
        "42:"  // Height 2: Multiply loop: Odd block loop
        "ldr s19, [x26], #0x4\n"
        "ldr s18, [x25], #0x4\n"
        "sub x27, x27, #0x1\n"
        "ldr q17, [x10, #0x0]\n"
        "ldr q16, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "fmla v20.4s, v17.4s, v19.s[0]\n"
        "fmla v22.4s, v17.4s, v18.s[0]\n"
        "fmla v21.4s, v16.4s, v19.s[0]\n"
        "fmla v23.4s, v16.4s, v18.s[0]\n"
        "cbnz x27, 42b\n"
        "43:"  // Height 2: Multiply loop: No odd multiplies
        "ldr w20, [%x[args_ptr], %[offsetof_num_strings]]\n"
        "add x28, x28, #0x1\n"
        "cmp x28, x20\n"
        "bne 36b\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "prfm pstl1keep, [x9, #0x0]\n"
        "add x26, x9, x20, LSL #2\n"
        "prfm pstl1keep, [x26, #0x0]\n"
        "tbz %x[flags], #1, 44f\n"
        "add x21, %x[args_ptr], %[offset_max]\n"
        "add x20, %x[args_ptr], %[offset_min]\n"
        "ld1r { v17.4s }, [x21]\n"
        "ld1r { v16.4s }, [x20]\n"
        "fmin v20.4s, v20.4s, v17.4s\n"
        "fmin v21.4s, v21.4s, v17.4s\n"
        "fmin v22.4s, v22.4s, v17.4s\n"
        "fmin v23.4s, v23.4s, v17.4s\n"
        "fmax v20.4s, v20.4s, v16.4s\n"
        "fmax v21.4s, v21.4s, v16.4s\n"
        "fmax v22.4s, v22.4s, v16.4s\n"
        "fmax v23.4s, v23.4s, v16.4s\n"
        "44:"  // Height 2: No activation
        "cmp x11, #0x8\n"
        "bge 49f\n"
        "tbz x11, #2, 46f\n"
        "st1 { v20.4s }, [x9], #0x10\n"
        "st1 { v22.4s }, [x26], #0x10\n"
        "tbz x11, #1, 45f\n"
        "str d21, [x9], #0x8\n"
        "str d23, [x26], #0x8\n"
        "tbz x11, #0, 48f\n"
        "st1 { v21.s }[2], [x9]\n"
        "st1 { v23.s }[2], [x26]\n"
        "b 48f\n"
        "45:"  // Height 2: Partial direct writeback: partial_1_4
        "tbz x11, #0, 48f\n"
        "str s21, [x9, #0x0]\n"
        "str s23, [x26, #0x0]\n"
        "b 48f\n"
        "46:"  // Height 2: Partial direct writeback: partial_2_0
        "tbz x11, #1, 47f\n"
        "str d20, [x9], #0x8\n"
        "str d22, [x26], #0x8\n"
        "tbz x11, #0, 48f\n"
        "st1 { v20.s }[2], [x9]\n"
        "st1 { v22.s }[2], [x26]\n"
        "b 48f\n"
        "47:"  // Height 2: Partial direct writeback: partial_1_0
        "str s20, [x9, #0x0]\n"
        "str s22, [x26, #0x0]\n"
        "48:"  // Height 2: Partial direct writeback: Done
        "b 50f\n"
        "49:"  // Height 2: Full writeback
        "str q20, [x9, #0x0]\n"
        "str q21, [x9, #0x10]\n"
        "add x9, x9, #0x20\n"
        "str q22, [x26, #0x0]\n"
        "str q23, [x26, #0x10]\n"
        "50:"  // Height 2: Writeback done
        "subs x11, x11, #0x8\n"
        "bgt 27b\n"
        "b 152f\n"
        "51:"  // Height 3
        "ldr x11, [%x[args_ptr], %[offsetof_N]]\n"
        "ldr x10, [%x[args_ptr], %[offsetof_B_ptr]]\n"
        "ldr x9, [%x[args_ptr], %[offsetof_output_ptr]]\n"
        "52:"  // Height 3: Column loop
        "cbz x10, 53f\n"
        "ldr q20, [x10, #0x0]\n"
        "ldr q21, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "mov v22.16b, v20.16b\n"
        "mov v23.16b, v21.16b\n"
        "mov v24.16b, v20.16b\n"
        "mov v25.16b, v21.16b\n"
        "b 60f\n"
        "53:"  // Height 3: no bias
        "tbz %x[flags], #0, 59f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "cmp x11, #0x8\n"
        "add x26, x9, x20, LSL #2\n"
        "add x25, x26, x20, LSL #2\n"
        "bge 58f\n"
        "tbz x11, #2, 55f\n"
        "ld1 { v20.4s }, [x9], #0x10\n"
        "ld1 { v22.4s }, [x26], #0x10\n"
        "ld1 { v24.4s }, [x25], #0x10\n"
        "tbz x11, #1, 54f\n"
        "ldr d21, [x9], #0x8\n"
        "ldr d23, [x26], #0x8\n"
        "mov x20, #0x18\n"
        "ldr d25, [x25], #0x8\n"
        "tbz x11, #0, 57f\n"
        "ld1 { v21.s }[2], [x9]\n"
        "ld1 { v23.s }[2], [x26]\n"
        "ld1 { v25.s }[2], [x25]\n"
        "b 57f\n"
        "54:"  // Height 3: Partial accumulate: partial_1_4
        "mov x20, #0x10\n"
        "tbz x11, #0, 57f\n"
        "ldr s21, [x9, #0x0]\n"
        "ldr s23, [x26, #0x0]\n"
        "ldr s25, [x25, #0x0]\n"
        "b 57f\n"
        "55:"  // Height 3: Partial accumulate: partial_2_0
        "tbz x11, #1, 56f\n"
        "ldr d20, [x9], #0x8\n"
        "ldr d22, [x26], #0x8\n"
        "mov x20, #0x8\n"
        "ldr d24, [x25], #0x8\n"
        "tbz x11, #0, 57f\n"
        "ld1 { v20.s }[2], [x9]\n"
        "ld1 { v22.s }[2], [x26]\n"
        "ld1 { v24.s }[2], [x25]\n"
        "b 57f\n"
        "56:"  // Height 3: Partial accumulate: partial_1_0
        "ldr s20, [x9, #0x0]\n"
        "ldr s22, [x26, #0x0]\n"
        "mov x20, #0x0\n"
        "ldr s24, [x25, #0x0]\n"
        "57:"  // Height 3: Partial accumulate: Done
        "sub x9, x9, x20\n"
        "b 60f\n"
        "58:"  // Height 3: full accumulate
        "ldr q20, [x9, #0x0]\n"
        "ldr q21, [x9, #0x10]\n"
        "ldr q22, [x26, #0x0]\n"
        "ldr q23, [x26, #0x10]\n"
        "ldr q24, [x25, #0x0]\n"
        "ldr q25, [x25, #0x10]\n"
        "b 60f\n"
        "59:"  // Height 3: no accumulate
        "movi v20.16b, #0x0\n"
        "movi v21.16b, #0x0\n"
        "movi v22.16b, #0x0\n"
        "movi v23.16b, #0x0\n"
        "movi v24.16b, #0x0\n"
        "movi v25.16b, #0x0\n"
        "60:"  // Height 3: setup done
        "mov x28, #0x0\n"
        "61:"  // Height 3: String loop
        "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
        "ldr x21, [%x[args_ptr], %[offsetof_input_offset]]\n"
        "ldr w27, [x20, x28, LSL #0x2]\n"
        "tbz %x[flags], #3, 62f\n"
        "ldr x20, [%x[input_ptr], x28, LSL #0x3]\n"
        "add x20, x20, x21, LSL #3\n"
        "ldr x26, [x20, #0x0]\n"
        "ldr x25, [x20, #0x8]\n"
        "ldr x24, [x20, #0x10]\n"
        "cbnz x28, 63f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
        "add x26, x26, x20, LSL #2\n"
        "add x25, x25, x20, LSL #2\n"
        "add x24, x24, x20, LSL #2\n"
        "b 63f\n"
        "62:"  // Height 3: setup direct input
        "mov x26, %x[input_ptr]\n"
        "add x25, x26, x21, LSL #2\n"
        "add x24, x25, x21, LSL #2\n"
        "63:"  // Height 3: input setup done
        "cmp x27, #0x4\n"
        "blt 66f\n"
        "ldr q0, [x26, #0x0]\n"
        "ldr q1, [x25, #0x0]\n"
        "cmp x27, #0x8\n"
        "ldr q2, [x24, #0x0]\n"
        "ldr q6, [x10, #0x0]\n"
        "ldr q7, [x10, #0x10]\n"
        "ldr q8, [x10, #0x20]\n"
        "ldr q9, [x10, #0x30]\n"
        "ldr q10, [x10, #0x40]\n"
        "ldr q11, [x10, #0x50]\n"
        "ldr q12, [x10, #0x60]\n"
        "ldr q13, [x10, #0x70]\n"
        "blt 65f\n"
        "64:"  // Height 3: Multiply loop: Main loop head
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "sub x27, x27, #0x4\n"
        "add x26, x26, #0x10\n"
        "fmla v24.4s, v6.4s, v2.s[0]\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "add x25, x25, #0x10\n"
        "add x24, x24, #0x10\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "fmla v25.4s, v7.4s, v2.s[0]\n"
        "cmp x27, #0x8\n"
        "add x10, x10, #0x80\n"
        "ldr q6, [x10, #0x0]\n"
        "ldr q7, [x10, #0x10]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "fmla v24.4s, v8.4s, v2.s[1]\n"
        "ldr q8, [x10, #0x20]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "prfm pldl1keep, [x24, #0x80]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "fmla v25.4s, v9.4s, v2.s[1]\n"
        "ldr q9, [x10, #0x30]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "fmla v24.4s, v10.4s, v2.s[2]\n"
        "ldr q10, [x10, #0x40]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "fmla v25.4s, v11.4s, v2.s[2]\n"
        "ldr q11, [x10, #0x50]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "fmla v24.4s, v12.4s, v2.s[3]\n"
        "ldr q12, [x10, #0x60]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "ldr q0, [x26, #0x0]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "ldr q1, [x25, #0x0]\n"
        "fmla v25.4s, v13.4s, v2.s[3]\n"
        "ldr q2, [x24, #0x0]\n"
        "ldr q13, [x10, #0x70]\n"
        "bge 64b\n"
        "65:"  // Height 3: Multiply loop: Single iteration only
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "add x26, x26, #0x10\n"
        "add x25, x25, #0x10\n"
        "fmla v24.4s, v6.4s, v2.s[0]\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "add x24, x24, #0x10\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "fmla v25.4s, v7.4s, v2.s[0]\n"
        "sub x27, x27, #0x4\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "add x10, x10, #0x80\n"
        "prfm pldl1keep, [x24, #0x80]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "fmla v24.4s, v8.4s, v2.s[1]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "fmla v25.4s, v9.4s, v2.s[1]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "fmla v24.4s, v10.4s, v2.s[2]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "fmla v25.4s, v11.4s, v2.s[2]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "fmla v24.4s, v12.4s, v2.s[3]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "fmla v25.4s, v13.4s, v2.s[3]\n"
        "66:"  // Height 3: Multiply loop: Main loop skip
        "cbz x27, 68f\n"
        "67:"  // Height 3: Multiply loop: Odd block loop
        "ldr s26, [x26], #0x4\n"
        "ldr s19, [x25], #0x4\n"
        "sub x27, x27, #0x1\n"
        "ldr s18, [x24], #0x4\n"
        "ldr q17, [x10, #0x0]\n"
        "ldr q16, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "fmla v20.4s, v17.4s, v26.s[0]\n"
        "fmla v22.4s, v17.4s, v19.s[0]\n"
        "fmla v24.4s, v17.4s, v18.s[0]\n"
        "fmla v21.4s, v16.4s, v26.s[0]\n"
        "fmla v23.4s, v16.4s, v19.s[0]\n"
        "fmla v25.4s, v16.4s, v18.s[0]\n"
        "cbnz x27, 67b\n"
        "68:"  // Height 3: Multiply loop: No odd multiplies
        "ldr w20, [%x[args_ptr], %[offsetof_num_strings]]\n"
        "add x28, x28, #0x1\n"
        "cmp x28, x20\n"
        "bne 61b\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "prfm pstl1keep, [x9, #0x0]\n"
        "add x26, x9, x20, LSL #2\n"
        "prfm pstl1keep, [x26, #0x0]\n"
        "add x25, x26, x20, LSL #2\n"
        "prfm pstl1keep, [x25, #0x0]\n"
        "tbz %x[flags], #1, 69f\n"
        "add x21, %x[args_ptr], %[offset_max]\n"
        "add x20, %x[args_ptr], %[offset_min]\n"
        "ld1r { v17.4s }, [x21]\n"
        "ld1r { v16.4s }, [x20]\n"
        "fmin v20.4s, v20.4s, v17.4s\n"
        "fmin v21.4s, v21.4s, v17.4s\n"
        "fmin v22.4s, v22.4s, v17.4s\n"
        "fmin v23.4s, v23.4s, v17.4s\n"
        "fmin v24.4s, v24.4s, v17.4s\n"
        "fmin v25.4s, v25.4s, v17.4s\n"
        "fmax v20.4s, v20.4s, v16.4s\n"
        "fmax v21.4s, v21.4s, v16.4s\n"
        "fmax v22.4s, v22.4s, v16.4s\n"
        "fmax v23.4s, v23.4s, v16.4s\n"
        "fmax v24.4s, v24.4s, v16.4s\n"
        "fmax v25.4s, v25.4s, v16.4s\n"
        "69:"  // Height 3: No activation
        "cmp x11, #0x8\n"
        "bge 74f\n"
        "tbz x11, #2, 71f\n"
        "st1 { v20.4s }, [x9], #0x10\n"
        "st1 { v22.4s }, [x26], #0x10\n"
        "st1 { v24.4s }, [x25], #0x10\n"
        "tbz x11, #1, 70f\n"
        "str d21, [x9], #0x8\n"
        "str d23, [x26], #0x8\n"
        "str d25, [x25], #0x8\n"
        "tbz x11, #0, 73f\n"
        "st1 { v21.s }[2], [x9]\n"
        "st1 { v23.s }[2], [x26]\n"
        "st1 { v25.s }[2], [x25]\n"
        "b 73f\n"
        "70:"  // Height 3: Partial direct writeback: partial_1_4
        "tbz x11, #0, 73f\n"
        "str s21, [x9, #0x0]\n"
        "str s23, [x26, #0x0]\n"
        "str s25, [x25, #0x0]\n"
        "b 73f\n"
        "71:"  // Height 3: Partial direct writeback: partial_2_0
        "tbz x11, #1, 72f\n"
        "str d20, [x9], #0x8\n"
        "str d22, [x26], #0x8\n"
        "str d24, [x25], #0x8\n"
        "tbz x11, #0, 73f\n"
        "st1 { v20.s }[2], [x9]\n"
        "st1 { v22.s }[2], [x26]\n"
        "st1 { v24.s }[2], [x25]\n"
        "b 73f\n"
        "72:"  // Height 3: Partial direct writeback: partial_1_0
        "str s20, [x9, #0x0]\n"
        "str s22, [x26, #0x0]\n"
        "str s24, [x25, #0x0]\n"
        "73:"  // Height 3: Partial direct writeback: Done
        "b 75f\n"
        "74:"  // Height 3: Full writeback
        "str q20, [x9, #0x0]\n"
        "str q21, [x9, #0x10]\n"
        "add x9, x9, #0x20\n"
        "str q22, [x26, #0x0]\n"
        "str q23, [x26, #0x10]\n"
        "str q24, [x25, #0x0]\n"
        "str q25, [x25, #0x10]\n"
        "75:"  // Height 3: Writeback done
        "subs x11, x11, #0x8\n"
        "bgt 52b\n"
        "b 152f\n"
        "76:"  // Height 4
        "ldr x11, [%x[args_ptr], %[offsetof_N]]\n"
        "ldr x10, [%x[args_ptr], %[offsetof_B_ptr]]\n"
        "ldr x9, [%x[args_ptr], %[offsetof_output_ptr]]\n"
        "77:"  // Height 4: Column loop
        "cbz x10, 78f\n"
        "ldr q20, [x10, #0x0]\n"
        "ldr q21, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "mov v22.16b, v20.16b\n"
        "mov v23.16b, v21.16b\n"
        "mov v24.16b, v20.16b\n"
        "mov v25.16b, v21.16b\n"
        "mov v26.16b, v20.16b\n"
        "mov v27.16b, v21.16b\n"
        "b 85f\n"
        "78:"  // Height 4: no bias
        "tbz %x[flags], #0, 84f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "cmp x11, #0x8\n"
        "add x26, x9, x20, LSL #2\n"
        "add x25, x26, x20, LSL #2\n"
        "add x24, x25, x20, LSL #2\n"
        "bge 83f\n"
        "tbz x11, #2, 80f\n"
        "ld1 { v20.4s }, [x9], #0x10\n"
        "ld1 { v22.4s }, [x26], #0x10\n"
        "ld1 { v24.4s }, [x25], #0x10\n"
        "ld1 { v26.4s }, [x24], #0x10\n"
        "tbz x11, #1, 79f\n"
        "ldr d21, [x9], #0x8\n"
        "ldr d23, [x26], #0x8\n"
        "mov x20, #0x18\n"
        "ldr d25, [x25], #0x8\n"
        "ldr d27, [x24], #0x8\n"
        "tbz x11, #0, 82f\n"
        "ld1 { v21.s }[2], [x9]\n"
        "ld1 { v23.s }[2], [x26]\n"
        "ld1 { v25.s }[2], [x25]\n"
        "ld1 { v27.s }[2], [x24]\n"
        "b 82f\n"
        "79:"  // Height 4: Partial accumulate: partial_1_4
        "mov x20, #0x10\n"
        "tbz x11, #0, 82f\n"
        "ldr s21, [x9, #0x0]\n"
        "ldr s23, [x26, #0x0]\n"
        "ldr s25, [x25, #0x0]\n"
        "ldr s27, [x24, #0x0]\n"
        "b 82f\n"
        "80:"  // Height 4: Partial accumulate: partial_2_0
        "tbz x11, #1, 81f\n"
        "ldr d20, [x9], #0x8\n"
        "ldr d22, [x26], #0x8\n"
        "mov x20, #0x8\n"
        "ldr d24, [x25], #0x8\n"
        "ldr d26, [x24], #0x8\n"
        "tbz x11, #0, 82f\n"
        "ld1 { v20.s }[2], [x9]\n"
        "ld1 { v22.s }[2], [x26]\n"
        "ld1 { v24.s }[2], [x25]\n"
        "ld1 { v26.s }[2], [x24]\n"
        "b 82f\n"
        "81:"  // Height 4: Partial accumulate: partial_1_0
        "ldr s20, [x9, #0x0]\n"
        "ldr s22, [x26, #0x0]\n"
        "mov x20, #0x0\n"
        "ldr s24, [x25, #0x0]\n"
        "ldr s26, [x24, #0x0]\n"
        "82:"  // Height 4: Partial accumulate: Done
        "sub x9, x9, x20\n"
        "b 85f\n"
        "83:"  // Height 4: full accumulate
        "ldr q20, [x9, #0x0]\n"
        "ldr q21, [x9, #0x10]\n"
        "ldr q22, [x26, #0x0]\n"
        "ldr q23, [x26, #0x10]\n"
        "ldr q24, [x25, #0x0]\n"
        "ldr q25, [x25, #0x10]\n"
        "ldr q26, [x24, #0x0]\n"
        "ldr q27, [x24, #0x10]\n"
        "b 85f\n"
        "84:"  // Height 4: no accumulate
        "movi v20.16b, #0x0\n"
        "movi v21.16b, #0x0\n"
        "movi v22.16b, #0x0\n"
        "movi v23.16b, #0x0\n"
        "movi v24.16b, #0x0\n"
        "movi v25.16b, #0x0\n"
        "movi v26.16b, #0x0\n"
        "movi v27.16b, #0x0\n"
        "85:"  // Height 4: setup done
        "mov x28, #0x0\n"
        "86:"  // Height 4: String loop
        "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
        "ldr x21, [%x[args_ptr], %[offsetof_input_offset]]\n"
        "ldr w27, [x20, x28, LSL #0x2]\n"
        "tbz %x[flags], #3, 87f\n"
        "ldr x20, [%x[input_ptr], x28, LSL #0x3]\n"
        "add x20, x20, x21, LSL #3\n"
        "ldr x26, [x20, #0x0]\n"
        "ldr x25, [x20, #0x8]\n"
        "ldr x24, [x20, #0x10]\n"
        "ldr x23, [x20, #0x18]\n"
        "cbnz x28, 88f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
        "add x26, x26, x20, LSL #2\n"
        "add x25, x25, x20, LSL #2\n"
        "add x24, x24, x20, LSL #2\n"
        "add x23, x23, x20, LSL #2\n"
        "b 88f\n"
        "87:"  // Height 4: setup direct input
        "mov x26, %x[input_ptr]\n"
        "add x25, x26, x21, LSL #2\n"
        "add x24, x25, x21, LSL #2\n"
        "add x23, x24, x21, LSL #2\n"
        "88:"  // Height 4: input setup done
        "cmp x27, #0x4\n"
        "blt 91f\n"
        "ldr q0, [x26, #0x0]\n"
        "ldr q1, [x25, #0x0]\n"
        "cmp x27, #0x8\n"
        "ldr q2, [x24, #0x0]\n"
        "ldr q3, [x23, #0x0]\n"
        "ldr q6, [x10, #0x0]\n"
        "ldr q7, [x10, #0x10]\n"
        "ldr q8, [x10, #0x20]\n"
        "ldr q9, [x10, #0x30]\n"
        "ldr q10, [x10, #0x40]\n"
        "ldr q11, [x10, #0x50]\n"
        "ldr q12, [x10, #0x60]\n"
        "ldr q13, [x10, #0x70]\n"
        "blt 90f\n"
        "89:"  // Height 4: Multiply loop: Main loop head
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "sub x27, x27, #0x4\n"
        "add x26, x26, #0x10\n"
        "fmla v24.4s, v6.4s, v2.s[0]\n"
        "fmla v26.4s, v6.4s, v3.s[0]\n"
        "add x25, x25, #0x10\n"
        "add x24, x24, #0x10\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "add x23, x23, #0x10\n"
        "cmp x27, #0x8\n"
        "fmla v25.4s, v7.4s, v2.s[0]\n"
        "fmla v27.4s, v7.4s, v3.s[0]\n"
        "add x10, x10, #0x80\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "ldr q6, [x10, #0x0]\n"
        "ldr q7, [x10, #0x10]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "fmla v24.4s, v8.4s, v2.s[1]\n"
        "fmla v26.4s, v8.4s, v3.s[1]\n"
        "ldr q8, [x10, #0x20]\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "prfm pldl1keep, [x24, #0x80]\n"
        "prfm pldl1keep, [x23, #0x80]\n"
        "fmla v25.4s, v9.4s, v2.s[1]\n"
        "fmla v27.4s, v9.4s, v3.s[1]\n"
        "ldr q9, [x10, #0x30]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "fmla v24.4s, v10.4s, v2.s[2]\n"
        "fmla v26.4s, v10.4s, v3.s[2]\n"
        "ldr q10, [x10, #0x40]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "fmla v25.4s, v11.4s, v2.s[2]\n"
        "fmla v27.4s, v11.4s, v3.s[2]\n"
        "ldr q11, [x10, #0x50]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "fmla v24.4s, v12.4s, v2.s[3]\n"
        "fmla v26.4s, v12.4s, v3.s[3]\n"
        "ldr q12, [x10, #0x60]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "ldr q0, [x26, #0x0]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "ldr q1, [x25, #0x0]\n"
        "fmla v25.4s, v13.4s, v2.s[3]\n"
        "ldr q2, [x24, #0x0]\n"
        "fmla v27.4s, v13.4s, v3.s[3]\n"
        "ldr q3, [x23, #0x0]\n"
        "ldr q13, [x10, #0x70]\n"
        "bge 89b\n"
        "90:"  // Height 4: Multiply loop: Single iteration only
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "add x26, x26, #0x10\n"
        "add x25, x25, #0x10\n"
        "fmla v24.4s, v6.4s, v2.s[0]\n"
        "fmla v26.4s, v6.4s, v3.s[0]\n"
        "add x24, x24, #0x10\n"
        "add x23, x23, #0x10\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "sub x27, x27, #0x4\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "fmla v25.4s, v7.4s, v2.s[0]\n"
        "fmla v27.4s, v7.4s, v3.s[0]\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "prfm pldl1keep, [x24, #0x80]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "prfm pldl1keep, [x23, #0x80]\n"
        "add x10, x10, #0x80\n"
        "fmla v24.4s, v8.4s, v2.s[1]\n"
        "fmla v26.4s, v8.4s, v3.s[1]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "fmla v25.4s, v9.4s, v2.s[1]\n"
        "fmla v27.4s, v9.4s, v3.s[1]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "fmla v24.4s, v10.4s, v2.s[2]\n"
        "fmla v26.4s, v10.4s, v3.s[2]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "fmla v25.4s, v11.4s, v2.s[2]\n"
        "fmla v27.4s, v11.4s, v3.s[2]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "fmla v24.4s, v12.4s, v2.s[3]\n"
        "fmla v26.4s, v12.4s, v3.s[3]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "fmla v25.4s, v13.4s, v2.s[3]\n"
        "fmla v27.4s, v13.4s, v3.s[3]\n"
        "91:"  // Height 4: Multiply loop: Main loop skip
        "cbz x27, 93f\n"
        "92:"  // Height 4: Multiply loop: Odd block loop
        "ldr s29, [x26], #0x4\n"
        "ldr s28, [x25], #0x4\n"
        "sub x27, x27, #0x1\n"
        "ldr s19, [x24], #0x4\n"
        "ldr s18, [x23], #0x4\n"
        "ldr q17, [x10, #0x0]\n"
        "ldr q16, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "fmla v20.4s, v17.4s, v29.s[0]\n"
        "fmla v22.4s, v17.4s, v28.s[0]\n"
        "fmla v24.4s, v17.4s, v19.s[0]\n"
        "fmla v26.4s, v17.4s, v18.s[0]\n"
        "fmla v21.4s, v16.4s, v29.s[0]\n"
        "fmla v23.4s, v16.4s, v28.s[0]\n"
        "fmla v25.4s, v16.4s, v19.s[0]\n"
        "fmla v27.4s, v16.4s, v18.s[0]\n"
        "cbnz x27, 92b\n"
        "93:"  // Height 4: Multiply loop: No odd multiplies
        "ldr w20, [%x[args_ptr], %[offsetof_num_strings]]\n"
        "add x28, x28, #0x1\n"
        "cmp x28, x20\n"
        "bne 86b\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "prfm pstl1keep, [x9, #0x0]\n"
        "add x26, x9, x20, LSL #2\n"
        "prfm pstl1keep, [x26, #0x0]\n"
        "add x25, x26, x20, LSL #2\n"
        "prfm pstl1keep, [x25, #0x0]\n"
        "add x24, x25, x20, LSL #2\n"
        "prfm pstl1keep, [x24, #0x0]\n"
        "tbz %x[flags], #1, 94f\n"
        "add x21, %x[args_ptr], %[offset_max]\n"
        "add x20, %x[args_ptr], %[offset_min]\n"
        "ld1r { v17.4s }, [x21]\n"
        "ld1r { v16.4s }, [x20]\n"
        "fmin v20.4s, v20.4s, v17.4s\n"
        "fmin v21.4s, v21.4s, v17.4s\n"
        "fmin v22.4s, v22.4s, v17.4s\n"
        "fmin v23.4s, v23.4s, v17.4s\n"
        "fmin v24.4s, v24.4s, v17.4s\n"
        "fmin v25.4s, v25.4s, v17.4s\n"
        "fmin v26.4s, v26.4s, v17.4s\n"
        "fmin v27.4s, v27.4s, v17.4s\n"
        "fmax v20.4s, v20.4s, v16.4s\n"
        "fmax v21.4s, v21.4s, v16.4s\n"
        "fmax v22.4s, v22.4s, v16.4s\n"
        "fmax v23.4s, v23.4s, v16.4s\n"
        "fmax v24.4s, v24.4s, v16.4s\n"
        "fmax v25.4s, v25.4s, v16.4s\n"
        "fmax v26.4s, v26.4s, v16.4s\n"
        "fmax v27.4s, v27.4s, v16.4s\n"
        "94:"  // Height 4: No activation
        "cmp x11, #0x8\n"
        "bge 99f\n"
        "tbz x11, #2, 96f\n"
        "st1 { v20.4s }, [x9], #0x10\n"
        "st1 { v22.4s }, [x26], #0x10\n"
        "st1 { v24.4s }, [x25], #0x10\n"
        "st1 { v26.4s }, [x24], #0x10\n"
        "tbz x11, #1, 95f\n"
        "str d21, [x9], #0x8\n"
        "str d23, [x26], #0x8\n"
        "str d25, [x25], #0x8\n"
        "str d27, [x24], #0x8\n"
        "tbz x11, #0, 98f\n"
        "st1 { v21.s }[2], [x9]\n"
        "st1 { v23.s }[2], [x26]\n"
        "st1 { v25.s }[2], [x25]\n"
        "st1 { v27.s }[2], [x24]\n"
        "b 98f\n"
        "95:"  // Height 4: Partial direct writeback: partial_1_4
        "tbz x11, #0, 98f\n"
        "str s21, [x9, #0x0]\n"
        "str s23, [x26, #0x0]\n"
        "str s25, [x25, #0x0]\n"
        "str s27, [x24, #0x0]\n"
        "b 98f\n"
        "96:"  // Height 4: Partial direct writeback: partial_2_0
        "tbz x11, #1, 97f\n"
        "str d20, [x9], #0x8\n"
        "str d22, [x26], #0x8\n"
        "str d24, [x25], #0x8\n"
        "str d26, [x24], #0x8\n"
        "tbz x11, #0, 98f\n"
        "st1 { v20.s }[2], [x9]\n"
        "st1 { v22.s }[2], [x26]\n"
        "st1 { v24.s }[2], [x25]\n"
        "st1 { v26.s }[2], [x24]\n"
        "b 98f\n"
        "97:"  // Height 4: Partial direct writeback: partial_1_0
        "str s20, [x9, #0x0]\n"
        "str s22, [x26, #0x0]\n"
        "str s24, [x25, #0x0]\n"
        "str s26, [x24, #0x0]\n"
        "98:"  // Height 4: Partial direct writeback: Done
        "b 100f\n"
        "99:"  // Height 4: Full writeback
        "str q20, [x9, #0x0]\n"
        "str q21, [x9, #0x10]\n"
        "add x9, x9, #0x20\n"
        "str q22, [x26, #0x0]\n"
        "str q23, [x26, #0x10]\n"
        "str q24, [x25, #0x0]\n"
        "str q25, [x25, #0x10]\n"
        "str q26, [x24, #0x0]\n"
        "str q27, [x24, #0x10]\n"
        "100:"  // Height 4: Writeback done
        "subs x11, x11, #0x8\n"
        "bgt 77b\n"
        "b 152f\n"
        "101:"  // Height 5
        "ldr x11, [%x[args_ptr], %[offsetof_N]]\n"
        "ldr x10, [%x[args_ptr], %[offsetof_B_ptr]]\n"
        "ldr x9, [%x[args_ptr], %[offsetof_output_ptr]]\n"
        "102:"  // Height 5: Column loop
        "cbz x10, 103f\n"
        "ldr q20, [x10, #0x0]\n"
        "ldr q21, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "mov v22.16b, v20.16b\n"
        "mov v23.16b, v21.16b\n"
        "mov v24.16b, v20.16b\n"
        "mov v25.16b, v21.16b\n"
        "mov v26.16b, v20.16b\n"
        "mov v27.16b, v21.16b\n"
        "mov v28.16b, v20.16b\n"
        "mov v29.16b, v21.16b\n"
        "b 110f\n"
        "103:"  // Height 5: no bias
        "tbz %x[flags], #0, 109f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "cmp x11, #0x8\n"
        "add x26, x9, x20, LSL #2\n"
        "add x25, x26, x20, LSL #2\n"
        "add x24, x25, x20, LSL #2\n"
        "add x23, x24, x20, LSL #2\n"
        "bge 108f\n"
        "tbz x11, #2, 105f\n"
        "ld1 { v20.4s }, [x9], #0x10\n"
        "ld1 { v22.4s }, [x26], #0x10\n"
        "ld1 { v24.4s }, [x25], #0x10\n"
        "ld1 { v26.4s }, [x24], #0x10\n"
        "ld1 { v28.4s }, [x23], #0x10\n"
        "tbz x11, #1, 104f\n"
        "ldr d21, [x9], #0x8\n"
        "ldr d23, [x26], #0x8\n"
        "mov x20, #0x18\n"
        "ldr d25, [x25], #0x8\n"
        "ldr d27, [x24], #0x8\n"
        "ldr d29, [x23], #0x8\n"
        "tbz x11, #0, 107f\n"
        "ld1 { v21.s }[2], [x9]\n"
        "ld1 { v23.s }[2], [x26]\n"
        "ld1 { v25.s }[2], [x25]\n"
        "ld1 { v27.s }[2], [x24]\n"
        "ld1 { v29.s }[2], [x23]\n"
        "b 107f\n"
        "104:"  // Height 5: Partial accumulate: partial_1_4
        "mov x20, #0x10\n"
        "tbz x11, #0, 107f\n"
        "ldr s21, [x9, #0x0]\n"
        "ldr s23, [x26, #0x0]\n"
        "ldr s25, [x25, #0x0]\n"
        "ldr s27, [x24, #0x0]\n"
        "ldr s29, [x23, #0x0]\n"
        "b 107f\n"
        "105:"  // Height 5: Partial accumulate: partial_2_0
        "tbz x11, #1, 106f\n"
        "ldr d20, [x9], #0x8\n"
        "ldr d22, [x26], #0x8\n"
        "mov x20, #0x8\n"
        "ldr d24, [x25], #0x8\n"
        "ldr d26, [x24], #0x8\n"
        "ldr d28, [x23], #0x8\n"
        "tbz x11, #0, 107f\n"
        "ld1 { v20.s }[2], [x9]\n"
        "ld1 { v22.s }[2], [x26]\n"
        "ld1 { v24.s }[2], [x25]\n"
        "ld1 { v26.s }[2], [x24]\n"
        "ld1 { v28.s }[2], [x23]\n"
        "b 107f\n"
        "106:"  // Height 5: Partial accumulate: partial_1_0
        "ldr s20, [x9, #0x0]\n"
        "ldr s22, [x26, #0x0]\n"
        "mov x20, #0x0\n"
        "ldr s24, [x25, #0x0]\n"
        "ldr s26, [x24, #0x0]\n"
        "ldr s28, [x23, #0x0]\n"
        "107:"  // Height 5: Partial accumulate: Done
        "sub x9, x9, x20\n"
        "b 110f\n"
        "108:"  // Height 5: full accumulate
        "ldr q20, [x9, #0x0]\n"
        "ldr q21, [x9, #0x10]\n"
        "ldr q22, [x26, #0x0]\n"
        "ldr q23, [x26, #0x10]\n"
        "ldr q24, [x25, #0x0]\n"
        "ldr q25, [x25, #0x10]\n"
        "ldr q26, [x24, #0x0]\n"
        "ldr q27, [x24, #0x10]\n"
        "ldr q28, [x23, #0x0]\n"
        "ldr q29, [x23, #0x10]\n"
        "b 110f\n"
        "109:"  // Height 5: no accumulate
        "movi v20.16b, #0x0\n"
        "movi v21.16b, #0x0\n"
        "movi v22.16b, #0x0\n"
        "movi v23.16b, #0x0\n"
        "movi v24.16b, #0x0\n"
        "movi v25.16b, #0x0\n"
        "movi v26.16b, #0x0\n"
        "movi v27.16b, #0x0\n"
        "movi v28.16b, #0x0\n"
        "movi v29.16b, #0x0\n"
        "110:"  // Height 5: setup done
        "mov x28, #0x0\n"
        "111:"  // Height 5: String loop
        "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
        "ldr x21, [%x[args_ptr], %[offsetof_input_offset]]\n"
        "ldr w27, [x20, x28, LSL #0x2]\n"
        "tbz %x[flags], #3, 112f\n"
        "ldr x20, [%x[input_ptr], x28, LSL #0x3]\n"
        "add x20, x20, x21, LSL #3\n"
        "ldr x26, [x20, #0x0]\n"
        "ldr x25, [x20, #0x8]\n"
        "ldr x24, [x20, #0x10]\n"
        "ldr x23, [x20, #0x18]\n"
        "ldr x22, [x20, #0x20]\n"
        "cbnz x28, 113f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
        "add x26, x26, x20, LSL #2\n"
        "add x25, x25, x20, LSL #2\n"
        "add x24, x24, x20, LSL #2\n"
        "add x23, x23, x20, LSL #2\n"
        "add x22, x22, x20, LSL #2\n"
        "b 113f\n"
        "112:"  // Height 5: setup direct input
        "mov x26, %x[input_ptr]\n"
        "add x25, x26, x21, LSL #2\n"
        "add x24, x25, x21, LSL #2\n"
        "add x23, x24, x21, LSL #2\n"
        "add x22, x23, x21, LSL #2\n"
        "113:"  // Height 5: input setup done
        "cmp x27, #0x4\n"
        "blt 116f\n"
        "ldr q0, [x26, #0x0]\n"
        "ldr q1, [x25, #0x0]\n"
        "cmp x27, #0x8\n"
        "ldr q2, [x24, #0x0]\n"
        "ldr q3, [x23, #0x0]\n"
        "ldr q4, [x22, #0x0]\n"
        "ldr q6, [x10, #0x0]\n"
        "ldr q7, [x10, #0x10]\n"
        "ldr q8, [x10, #0x20]\n"
        "ldr q9, [x10, #0x30]\n"
        "ldr q10, [x10, #0x40]\n"
        "ldr q11, [x10, #0x50]\n"
        "ldr q12, [x10, #0x60]\n"
        "ldr q13, [x10, #0x70]\n"
        "blt 115f\n"
        "114:"  // Height 5: Multiply loop: Main loop head
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "sub x27, x27, #0x4\n"
        "add x26, x26, #0x10\n"
        "fmla v24.4s, v6.4s, v2.s[0]\n"
        "fmla v26.4s, v6.4s, v3.s[0]\n"
        "add x25, x25, #0x10\n"
        "add x24, x24, #0x10\n"
        "fmla v28.4s, v6.4s, v4.s[0]\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "add x23, x23, #0x10\n"
        "add x22, x22, #0x10\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "fmla v25.4s, v7.4s, v2.s[0]\n"
        "cmp x27, #0x8\n"
        "add x10, x10, #0x80\n"
        "ldr q6, [x10, #0x0]\n"
        "fmla v27.4s, v7.4s, v3.s[0]\n"
        "fmla v29.4s, v7.4s, v4.s[0]\n"
        "ldr q7, [x10, #0x10]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "fmla v24.4s, v8.4s, v2.s[1]\n"
        "fmla v26.4s, v8.4s, v3.s[1]\n"
        "prfm pldl1keep, [x24, #0x80]\n"
        "prfm pldl1keep, [x23, #0x80]\n"
        "fmla v28.4s, v8.4s, v4.s[1]\n"
        "ldr q8, [x10, #0x20]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "prfm pldl1keep, [x22, #0x80]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "fmla v25.4s, v9.4s, v2.s[1]\n"
        "fmla v27.4s, v9.4s, v3.s[1]\n"
        "fmla v29.4s, v9.4s, v4.s[1]\n"
        "ldr q9, [x10, #0x30]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "fmla v24.4s, v10.4s, v2.s[2]\n"
        "fmla v26.4s, v10.4s, v3.s[2]\n"
        "fmla v28.4s, v10.4s, v4.s[2]\n"
        "ldr q10, [x10, #0x40]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "fmla v25.4s, v11.4s, v2.s[2]\n"
        "fmla v27.4s, v11.4s, v3.s[2]\n"
        "fmla v29.4s, v11.4s, v4.s[2]\n"
        "ldr q11, [x10, #0x50]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "fmla v24.4s, v12.4s, v2.s[3]\n"
        "fmla v26.4s, v12.4s, v3.s[3]\n"
        "fmla v28.4s, v12.4s, v4.s[3]\n"
        "ldr q12, [x10, #0x60]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "ldr q0, [x26, #0x0]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "ldr q1, [x25, #0x0]\n"
        "fmla v25.4s, v13.4s, v2.s[3]\n"
        "ldr q2, [x24, #0x0]\n"
        "fmla v27.4s, v13.4s, v3.s[3]\n"
        "ldr q3, [x23, #0x0]\n"
        "fmla v29.4s, v13.4s, v4.s[3]\n"
        "ldr q4, [x22, #0x0]\n"
        "ldr q13, [x10, #0x70]\n"
        "bge 114b\n"
        "115:"  // Height 5: Multiply loop: Single iteration only
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "add x26, x26, #0x10\n"
        "add x25, x25, #0x10\n"
        "fmla v24.4s, v6.4s, v2.s[0]\n"
        "fmla v26.4s, v6.4s, v3.s[0]\n"
        "add x24, x24, #0x10\n"
        "add x23, x23, #0x10\n"
        "fmla v28.4s, v6.4s, v4.s[0]\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "add x22, x22, #0x10\n"
        "sub x27, x27, #0x4\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "fmla v25.4s, v7.4s, v2.s[0]\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "fmla v27.4s, v7.4s, v3.s[0]\n"
        "fmla v29.4s, v7.4s, v4.s[0]\n"
        "prfm pldl1keep, [x24, #0x80]\n"
        "prfm pldl1keep, [x23, #0x80]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "prfm pldl1keep, [x22, #0x80]\n"
        "add x10, x10, #0x80\n"
        "fmla v24.4s, v8.4s, v2.s[1]\n"
        "fmla v26.4s, v8.4s, v3.s[1]\n"
        "fmla v28.4s, v8.4s, v4.s[1]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "fmla v25.4s, v9.4s, v2.s[1]\n"
        "fmla v27.4s, v9.4s, v3.s[1]\n"
        "fmla v29.4s, v9.4s, v4.s[1]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "fmla v24.4s, v10.4s, v2.s[2]\n"
        "fmla v26.4s, v10.4s, v3.s[2]\n"
        "fmla v28.4s, v10.4s, v4.s[2]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "fmla v25.4s, v11.4s, v2.s[2]\n"
        "fmla v27.4s, v11.4s, v3.s[2]\n"
        "fmla v29.4s, v11.4s, v4.s[2]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "fmla v24.4s, v12.4s, v2.s[3]\n"
        "fmla v26.4s, v12.4s, v3.s[3]\n"
        "fmla v28.4s, v12.4s, v4.s[3]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "fmla v25.4s, v13.4s, v2.s[3]\n"
        "fmla v27.4s, v13.4s, v3.s[3]\n"
        "fmla v29.4s, v13.4s, v4.s[3]\n"
        "116:"  // Height 5: Multiply loop: Main loop skip
        "cbz x27, 118f\n"
        "117:"  // Height 5: Multiply loop: Odd block loop
        "ldr s0, [x26], #0x4\n"
        "ldr s31, [x25], #0x4\n"
        "sub x27, x27, #0x1\n"
        "ldr s30, [x24], #0x4\n"
        "ldr s19, [x23], #0x4\n"
        "ldr s18, [x22], #0x4\n"
        "ldr q17, [x10, #0x0]\n"
        "ldr q16, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "fmla v20.4s, v17.4s, v0.s[0]\n"
        "fmla v22.4s, v17.4s, v31.s[0]\n"
        "fmla v24.4s, v17.4s, v30.s[0]\n"
        "fmla v26.4s, v17.4s, v19.s[0]\n"
        "fmla v28.4s, v17.4s, v18.s[0]\n"
        "fmla v21.4s, v16.4s, v0.s[0]\n"
        "fmla v23.4s, v16.4s, v31.s[0]\n"
        "fmla v25.4s, v16.4s, v30.s[0]\n"
        "fmla v27.4s, v16.4s, v19.s[0]\n"
        "fmla v29.4s, v16.4s, v18.s[0]\n"
        "cbnz x27, 117b\n"
        "118:"  // Height 5: Multiply loop: No odd multiplies
        "ldr w20, [%x[args_ptr], %[offsetof_num_strings]]\n"
        "add x28, x28, #0x1\n"
        "cmp x28, x20\n"
        "bne 111b\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "prfm pstl1keep, [x9, #0x0]\n"
        "add x26, x9, x20, LSL #2\n"
        "prfm pstl1keep, [x26, #0x0]\n"
        "add x25, x26, x20, LSL #2\n"
        "prfm pstl1keep, [x25, #0x0]\n"
        "add x24, x25, x20, LSL #2\n"
        "prfm pstl1keep, [x24, #0x0]\n"
        "add x23, x24, x20, LSL #2\n"
        "prfm pstl1keep, [x23, #0x0]\n"
        "tbz %x[flags], #1, 119f\n"
        "add x21, %x[args_ptr], %[offset_max]\n"
        "add x20, %x[args_ptr], %[offset_min]\n"
        "ld1r { v17.4s }, [x21]\n"
        "ld1r { v16.4s }, [x20]\n"
        "fmin v20.4s, v20.4s, v17.4s\n"
        "fmin v21.4s, v21.4s, v17.4s\n"
        "fmin v22.4s, v22.4s, v17.4s\n"
        "fmin v23.4s, v23.4s, v17.4s\n"
        "fmin v24.4s, v24.4s, v17.4s\n"
        "fmin v25.4s, v25.4s, v17.4s\n"
        "fmin v26.4s, v26.4s, v17.4s\n"
        "fmin v27.4s, v27.4s, v17.4s\n"
        "fmin v28.4s, v28.4s, v17.4s\n"
        "fmin v29.4s, v29.4s, v17.4s\n"
        "fmax v20.4s, v20.4s, v16.4s\n"
        "fmax v21.4s, v21.4s, v16.4s\n"
        "fmax v22.4s, v22.4s, v16.4s\n"
        "fmax v23.4s, v23.4s, v16.4s\n"
        "fmax v24.4s, v24.4s, v16.4s\n"
        "fmax v25.4s, v25.4s, v16.4s\n"
        "fmax v26.4s, v26.4s, v16.4s\n"
        "fmax v27.4s, v27.4s, v16.4s\n"
        "fmax v28.4s, v28.4s, v16.4s\n"
        "fmax v29.4s, v29.4s, v16.4s\n"
        "119:"  // Height 5: No activation
        "cmp x11, #0x8\n"
        "bge 124f\n"
        "tbz x11, #2, 121f\n"
        "st1 { v20.4s }, [x9], #0x10\n"
        "st1 { v22.4s }, [x26], #0x10\n"
        "st1 { v24.4s }, [x25], #0x10\n"
        "st1 { v26.4s }, [x24], #0x10\n"
        "st1 { v28.4s }, [x23], #0x10\n"
        "tbz x11, #1, 120f\n"
        "str d21, [x9], #0x8\n"
        "str d23, [x26], #0x8\n"
        "str d25, [x25], #0x8\n"
        "str d27, [x24], #0x8\n"
        "str d29, [x23], #0x8\n"
        "tbz x11, #0, 123f\n"
        "st1 { v21.s }[2], [x9]\n"
        "st1 { v23.s }[2], [x26]\n"
        "st1 { v25.s }[2], [x25]\n"
        "st1 { v27.s }[2], [x24]\n"
        "st1 { v29.s }[2], [x23]\n"
        "b 123f\n"
        "120:"  // Height 5: Partial direct writeback: partial_1_4
        "tbz x11, #0, 123f\n"
        "str s21, [x9, #0x0]\n"
        "str s23, [x26, #0x0]\n"
        "str s25, [x25, #0x0]\n"
        "str s27, [x24, #0x0]\n"
        "str s29, [x23, #0x0]\n"
        "b 123f\n"
        "121:"  // Height 5: Partial direct writeback: partial_2_0
        "tbz x11, #1, 122f\n"
        "str d20, [x9], #0x8\n"
        "str d22, [x26], #0x8\n"
        "str d24, [x25], #0x8\n"
        "str d26, [x24], #0x8\n"
        "str d28, [x23], #0x8\n"
        "tbz x11, #0, 123f\n"
        "st1 { v20.s }[2], [x9]\n"
        "st1 { v22.s }[2], [x26]\n"
        "st1 { v24.s }[2], [x25]\n"
        "st1 { v26.s }[2], [x24]\n"
        "st1 { v28.s }[2], [x23]\n"
        "b 123f\n"
        "122:"  // Height 5: Partial direct writeback: partial_1_0
        "str s20, [x9, #0x0]\n"
        "str s22, [x26, #0x0]\n"
        "str s24, [x25, #0x0]\n"
        "str s26, [x24, #0x0]\n"
        "str s28, [x23, #0x0]\n"
        "123:"  // Height 5: Partial direct writeback: Done
        "b 125f\n"
        "124:"  // Height 5: Full writeback
        "str q20, [x9, #0x0]\n"
        "str q21, [x9, #0x10]\n"
        "add x9, x9, #0x20\n"
        "str q22, [x26, #0x0]\n"
        "str q23, [x26, #0x10]\n"
        "str q24, [x25, #0x0]\n"
        "str q25, [x25, #0x10]\n"
        "str q26, [x24, #0x0]\n"
        "str q27, [x24, #0x10]\n"
        "str q28, [x23, #0x0]\n"
        "str q29, [x23, #0x10]\n"
        "125:"  // Height 5: Writeback done
        "subs x11, x11, #0x8\n"
        "bgt 102b\n"
        "b 152f\n"
        "126:"  // Height 6
        "ldr x21, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "ldr x9, [%x[args_ptr], %[offsetof_output_ptr]]\n"
        "mov x20, #0x18\n"
        "ldr x11, [%x[args_ptr], %[offsetof_N]]\n"
        "ldr x10, [%x[args_ptr], %[offsetof_B_ptr]]\n"
        "madd x20, x21, x20, x9\n"
        "str x20, [%x[args_ptr], %[offsetof_output_ptr]]\n"
        "127:"  // Height 6: Column loop
        "cbz x10, 128f\n"
        "ldr q20, [x10, #0x0]\n"
        "ldr q21, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "mov v22.16b, v20.16b\n"
        "mov v23.16b, v21.16b\n"
        "mov v24.16b, v20.16b\n"
        "mov v25.16b, v21.16b\n"
        "mov v26.16b, v20.16b\n"
        "mov v27.16b, v21.16b\n"
        "mov v28.16b, v20.16b\n"
        "mov v29.16b, v21.16b\n"
        "mov v30.16b, v20.16b\n"
        "mov v31.16b, v21.16b\n"
        "b 135f\n"
        "128:"  // Height 6: no bias
        "tbz %x[flags], #0, 134f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "cmp x11, #0x8\n"
        "add x26, x9, x20, LSL #2\n"
        "add x25, x26, x20, LSL #2\n"
        "add x24, x25, x20, LSL #2\n"
        "add x23, x24, x20, LSL #2\n"
        "add x22, x23, x20, LSL #2\n"
        "bge 133f\n"
        "tbz x11, #2, 130f\n"
        "ld1 { v20.4s }, [x9], #0x10\n"
        "ld1 { v22.4s }, [x26], #0x10\n"
        "ld1 { v24.4s }, [x25], #0x10\n"
        "ld1 { v26.4s }, [x24], #0x10\n"
        "ld1 { v28.4s }, [x23], #0x10\n"
        "ld1 { v30.4s }, [x22], #0x10\n"
        "tbz x11, #1, 129f\n"
        "ldr d21, [x9], #0x8\n"
        "ldr d23, [x26], #0x8\n"
        "mov x20, #0x18\n"
        "ldr d25, [x25], #0x8\n"
        "ldr d27, [x24], #0x8\n"
        "ldr d29, [x23], #0x8\n"
        "ldr d31, [x22], #0x8\n"
        "tbz x11, #0, 132f\n"
        "ld1 { v21.s }[2], [x9]\n"
        "ld1 { v23.s }[2], [x26]\n"
        "ld1 { v25.s }[2], [x25]\n"
        "ld1 { v27.s }[2], [x24]\n"
        "ld1 { v29.s }[2], [x23]\n"
        "ld1 { v31.s }[2], [x22]\n"
        "b 132f\n"
        "129:"  // Height 6: Partial accumulate: partial_1_4
        "mov x20, #0x10\n"
        "tbz x11, #0, 132f\n"
        "ldr s21, [x9, #0x0]\n"
        "ldr s23, [x26, #0x0]\n"
        "ldr s25, [x25, #0x0]\n"
        "ldr s27, [x24, #0x0]\n"
        "ldr s29, [x23, #0x0]\n"
        "ldr s31, [x22, #0x0]\n"
        "b 132f\n"
        "130:"  // Height 6: Partial accumulate: partial_2_0
        "tbz x11, #1, 131f\n"
        "ldr d20, [x9], #0x8\n"
        "ldr d22, [x26], #0x8\n"
        "mov x20, #0x8\n"
        "ldr d24, [x25], #0x8\n"
        "ldr d26, [x24], #0x8\n"
        "ldr d28, [x23], #0x8\n"
        "ldr d30, [x22], #0x8\n"
        "tbz x11, #0, 132f\n"
        "ld1 { v20.s }[2], [x9]\n"
        "ld1 { v22.s }[2], [x26]\n"
        "ld1 { v24.s }[2], [x25]\n"
        "ld1 { v26.s }[2], [x24]\n"
        "ld1 { v28.s }[2], [x23]\n"
        "ld1 { v30.s }[2], [x22]\n"
        "b 132f\n"
        "131:"  // Height 6: Partial accumulate: partial_1_0
        "ldr s20, [x9, #0x0]\n"
        "ldr s22, [x26, #0x0]\n"
        "mov x20, #0x0\n"
        "ldr s24, [x25, #0x0]\n"
        "ldr s26, [x24, #0x0]\n"
        "ldr s28, [x23, #0x0]\n"
        "ldr s30, [x22, #0x0]\n"
        "132:"  // Height 6: Partial accumulate: Done
        "sub x9, x9, x20\n"
        "b 135f\n"
        "133:"  // Height 6: full accumulate
        "ldr q20, [x9, #0x0]\n"
        "ldr q21, [x9, #0x10]\n"
        "ldr q22, [x26, #0x0]\n"
        "ldr q23, [x26, #0x10]\n"
        "ldr q24, [x25, #0x0]\n"
        "ldr q25, [x25, #0x10]\n"
        "ldr q26, [x24, #0x0]\n"
        "ldr q27, [x24, #0x10]\n"
        "ldr q28, [x23, #0x0]\n"
        "ldr q29, [x23, #0x10]\n"
        "ldr q30, [x22, #0x0]\n"
        "ldr q31, [x22, #0x10]\n"
        "b 135f\n"
        "134:"  // Height 6: no accumulate
        "movi v20.16b, #0x0\n"
        "movi v21.16b, #0x0\n"
        "movi v22.16b, #0x0\n"
        "movi v23.16b, #0x0\n"
        "movi v24.16b, #0x0\n"
        "movi v25.16b, #0x0\n"
        "movi v26.16b, #0x0\n"
        "movi v27.16b, #0x0\n"
        "movi v28.16b, #0x0\n"
        "movi v29.16b, #0x0\n"
        "movi v30.16b, #0x0\n"
        "movi v31.16b, #0x0\n"
        "135:"  // Height 6: setup done
        "mov x28, #0x0\n"
        "136:"  // Height 6: String loop
        "ldr x20, [%x[args_ptr], %[offsetof_string_lengths]]\n"
        "ldr x21, [%x[args_ptr], %[offsetof_input_offset]]\n"
        "ldr w27, [x20, x28, LSL #0x2]\n"
        "tbz %x[flags], #3, 137f\n"
        "ldr x20, [%x[input_ptr], x28, LSL #0x3]\n"
        "add x20, x20, x21, LSL #3\n"
        "ldr x26, [x20, #0x0]\n"
        "ldr x25, [x20, #0x8]\n"
        "ldr x24, [x20, #0x10]\n"
        "ldr x23, [x20, #0x18]\n"
        "ldr x22, [x20, #0x20]\n"
        "ldr x21, [x20, #0x28]\n"
        "cbnz x28, 138f\n"
        "ldr x20, [%x[args_ptr], %[offsetof_input_initial_col]]\n"
        "add x26, x26, x20, LSL #2\n"
        "add x25, x25, x20, LSL #2\n"
        "add x24, x24, x20, LSL #2\n"
        "add x23, x23, x20, LSL #2\n"
        "add x22, x22, x20, LSL #2\n"
        "add x21, x21, x20, LSL #2\n"
        "b 138f\n"
        "137:"  // Height 6: setup direct input
        "mov x26, %x[input_ptr]\n"
        "add x25, x26, x21, LSL #2\n"
        "add x24, x25, x21, LSL #2\n"
        "add x23, x24, x21, LSL #2\n"
        "add x22, x23, x21, LSL #2\n"
        "add x21, x22, x21, LSL #2\n"
        "138:"  // Height 6: input setup done
        "cmp x27, #0x4\n"
        "blt 141f\n"
        "ldr q0, [x26, #0x0]\n"
        "ldr q1, [x25, #0x0]\n"
        "cmp x27, #0x8\n"
        "ldr q2, [x24, #0x0]\n"
        "ldr q3, [x23, #0x0]\n"
        "ldr q4, [x22, #0x0]\n"
        "ldr q5, [x21, #0x0]\n"
        "ldr q6, [x10, #0x0]\n"
        "ldr q7, [x10, #0x10]\n"
        "ldr q8, [x10, #0x20]\n"
        "ldr q9, [x10, #0x30]\n"
        "ldr q10, [x10, #0x40]\n"
        "ldr q11, [x10, #0x50]\n"
        "ldr q12, [x10, #0x60]\n"
        "ldr q13, [x10, #0x70]\n"
        "blt 140f\n"
        "139:"  // Height 6: Multiply loop: Main loop head
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "sub x27, x27, #0x4\n"
        "add x26, x26, #0x10\n"
        "fmla v24.4s, v6.4s, v2.s[0]\n"
        "fmla v26.4s, v6.4s, v3.s[0]\n"
        "add x25, x25, #0x10\n"
        "add x24, x24, #0x10\n"
        "fmla v28.4s, v6.4s, v4.s[0]\n"
        "fmla v30.4s, v6.4s, v5.s[0]\n"
        "add x23, x23, #0x10\n"
        "add x22, x22, #0x10\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "add x21, x21, #0x10\n"
        "cmp x27, #0x8\n"
        "fmla v25.4s, v7.4s, v2.s[0]\n"
        "fmla v27.4s, v7.4s, v3.s[0]\n"
        "add x10, x10, #0x80\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "ldr q6, [x10, #0x0]\n"
        "fmla v29.4s, v7.4s, v4.s[0]\n"
        "fmla v31.4s, v7.4s, v5.s[0]\n"
        "ldr q7, [x10, #0x10]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "prfm pldl1keep, [x24, #0x80]\n"
        "fmla v24.4s, v8.4s, v2.s[1]\n"
        "fmla v26.4s, v8.4s, v3.s[1]\n"
        "prfm pldl1keep, [x23, #0x80]\n"
        "prfm pldl1keep, [x22, #0x80]\n"
        "fmla v28.4s, v8.4s, v4.s[1]\n"
        "fmla v30.4s, v8.4s, v5.s[1]\n"
        "ldr q8, [x10, #0x20]\n"
        "prfm pldl1keep, [x21, #0x80]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "fmla v25.4s, v9.4s, v2.s[1]\n"
        "fmla v27.4s, v9.4s, v3.s[1]\n"
        "fmla v29.4s, v9.4s, v4.s[1]\n"
        "fmla v31.4s, v9.4s, v5.s[1]\n"
        "ldr q9, [x10, #0x30]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "fmla v24.4s, v10.4s, v2.s[2]\n"
        "fmla v26.4s, v10.4s, v3.s[2]\n"
        "fmla v28.4s, v10.4s, v4.s[2]\n"
        "fmla v30.4s, v10.4s, v5.s[2]\n"
        "ldr q10, [x10, #0x40]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "fmla v25.4s, v11.4s, v2.s[2]\n"
        "fmla v27.4s, v11.4s, v3.s[2]\n"
        "fmla v29.4s, v11.4s, v4.s[2]\n"
        "fmla v31.4s, v11.4s, v5.s[2]\n"
        "ldr q11, [x10, #0x50]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "fmla v24.4s, v12.4s, v2.s[3]\n"
        "fmla v26.4s, v12.4s, v3.s[3]\n"
        "fmla v28.4s, v12.4s, v4.s[3]\n"
        "fmla v30.4s, v12.4s, v5.s[3]\n"
        "ldr q12, [x10, #0x60]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "ldr q0, [x26, #0x0]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "ldr q1, [x25, #0x0]\n"
        "fmla v25.4s, v13.4s, v2.s[3]\n"
        "ldr q2, [x24, #0x0]\n"
        "fmla v27.4s, v13.4s, v3.s[3]\n"
        "ldr q3, [x23, #0x0]\n"
        "fmla v29.4s, v13.4s, v4.s[3]\n"
        "ldr q4, [x22, #0x0]\n"
        "fmla v31.4s, v13.4s, v5.s[3]\n"
        "ldr q5, [x21, #0x0]\n"
        "ldr q13, [x10, #0x70]\n"
        "bge 139b\n"
        "140:"  // Height 6: Multiply loop: Single iteration only
        "fmla v20.4s, v6.4s, v0.s[0]\n"
        "fmla v22.4s, v6.4s, v1.s[0]\n"
        "add x26, x26, #0x10\n"
        "add x25, x25, #0x10\n"
        "fmla v24.4s, v6.4s, v2.s[0]\n"
        "fmla v26.4s, v6.4s, v3.s[0]\n"
        "add x24, x24, #0x10\n"
        "add x23, x23, #0x10\n"
        "fmla v28.4s, v6.4s, v4.s[0]\n"
        "fmla v30.4s, v6.4s, v5.s[0]\n"
        "add x22, x22, #0x10\n"
        "add x21, x21, #0x10\n"
        "fmla v21.4s, v7.4s, v0.s[0]\n"
        "fmla v23.4s, v7.4s, v1.s[0]\n"
        "sub x27, x27, #0x4\n"
        "prfm pldl1keep, [x26, #0x80]\n"
        "fmla v25.4s, v7.4s, v2.s[0]\n"
        "fmla v27.4s, v7.4s, v3.s[0]\n"
        "prfm pldl1keep, [x25, #0x80]\n"
        "prfm pldl1keep, [x24, #0x80]\n"
        "fmla v29.4s, v7.4s, v4.s[0]\n"
        "fmla v31.4s, v7.4s, v5.s[0]\n"
        "prfm pldl1keep, [x23, #0x80]\n"
        "prfm pldl1keep, [x22, #0x80]\n"
        "fmla v20.4s, v8.4s, v0.s[1]\n"
        "fmla v22.4s, v8.4s, v1.s[1]\n"
        "prfm pldl1keep, [x21, #0x80]\n"
        "add x10, x10, #0x80\n"
        "fmla v24.4s, v8.4s, v2.s[1]\n"
        "fmla v26.4s, v8.4s, v3.s[1]\n"
        "fmla v28.4s, v8.4s, v4.s[1]\n"
        "fmla v30.4s, v8.4s, v5.s[1]\n"
        "fmla v21.4s, v9.4s, v0.s[1]\n"
        "fmla v23.4s, v9.4s, v1.s[1]\n"
        "fmla v25.4s, v9.4s, v2.s[1]\n"
        "fmla v27.4s, v9.4s, v3.s[1]\n"
        "fmla v29.4s, v9.4s, v4.s[1]\n"
        "fmla v31.4s, v9.4s, v5.s[1]\n"
        "fmla v20.4s, v10.4s, v0.s[2]\n"
        "fmla v22.4s, v10.4s, v1.s[2]\n"
        "fmla v24.4s, v10.4s, v2.s[2]\n"
        "fmla v26.4s, v10.4s, v3.s[2]\n"
        "fmla v28.4s, v10.4s, v4.s[2]\n"
        "fmla v30.4s, v10.4s, v5.s[2]\n"
        "fmla v21.4s, v11.4s, v0.s[2]\n"
        "fmla v23.4s, v11.4s, v1.s[2]\n"
        "fmla v25.4s, v11.4s, v2.s[2]\n"
        "fmla v27.4s, v11.4s, v3.s[2]\n"
        "fmla v29.4s, v11.4s, v4.s[2]\n"
        "fmla v31.4s, v11.4s, v5.s[2]\n"
        "fmla v20.4s, v12.4s, v0.s[3]\n"
        "fmla v22.4s, v12.4s, v1.s[3]\n"
        "fmla v24.4s, v12.4s, v2.s[3]\n"
        "fmla v26.4s, v12.4s, v3.s[3]\n"
        "fmla v28.4s, v12.4s, v4.s[3]\n"
        "fmla v30.4s, v12.4s, v5.s[3]\n"
        "fmla v21.4s, v13.4s, v0.s[3]\n"
        "fmla v23.4s, v13.4s, v1.s[3]\n"
        "fmla v25.4s, v13.4s, v2.s[3]\n"
        "fmla v27.4s, v13.4s, v3.s[3]\n"
        "fmla v29.4s, v13.4s, v4.s[3]\n"
        "fmla v31.4s, v13.4s, v5.s[3]\n"
        "141:"  // Height 6: Multiply loop: Main loop skip
        "cbz x27, 143f\n"
        "142:"  // Height 6: Multiply loop: Odd block loop
        "ldr s3, [x26], #0x4\n"
        "ldr s2, [x25], #0x4\n"
        "sub x27, x27, #0x1\n"
        "ldr s1, [x24], #0x4\n"
        "ldr s0, [x23], #0x4\n"
        "ldr s19, [x22], #0x4\n"
        "ldr s18, [x21], #0x4\n"
        "ldr q17, [x10, #0x0]\n"
        "ldr q16, [x10, #0x10]\n"
        "add x10, x10, #0x20\n"
        "fmla v20.4s, v17.4s, v3.s[0]\n"
        "fmla v22.4s, v17.4s, v2.s[0]\n"
        "fmla v24.4s, v17.4s, v1.s[0]\n"
        "fmla v26.4s, v17.4s, v0.s[0]\n"
        "fmla v28.4s, v17.4s, v19.s[0]\n"
        "fmla v30.4s, v17.4s, v18.s[0]\n"
        "fmla v21.4s, v16.4s, v3.s[0]\n"
        "fmla v23.4s, v16.4s, v2.s[0]\n"
        "fmla v25.4s, v16.4s, v1.s[0]\n"
        "fmla v27.4s, v16.4s, v0.s[0]\n"
        "fmla v29.4s, v16.4s, v19.s[0]\n"
        "fmla v31.4s, v16.4s, v18.s[0]\n"
        "cbnz x27, 142b\n"
        "143:"  // Height 6: Multiply loop: No odd multiplies
        "ldr w20, [%x[args_ptr], %[offsetof_num_strings]]\n"
        "add x28, x28, #0x1\n"
        "cmp x28, x20\n"
        "bne 136b\n"
        "ldr x20, [%x[args_ptr], %[offsetof_output_offset]]\n"
        "prfm pstl1keep, [x9, #0x0]\n"
        "add x26, x9, x20, LSL #2\n"
        "prfm pstl1keep, [x26, #0x0]\n"
        "add x25, x26, x20, LSL #2\n"
        "prfm pstl1keep, [x25, #0x0]\n"
        "add x24, x25, x20, LSL #2\n"
        "prfm pstl1keep, [x24, #0x0]\n"
        "add x23, x24, x20, LSL #2\n"
        "add x22, x23, x20, LSL #2\n"
        "prfm pstl1keep, [x23, #0x0]\n"
        "prfm pstl1keep, [x22, #0x0]\n"
        "tbz %x[flags], #1, 144f\n"
        "add x21, %x[args_ptr], %[offset_max]\n"
        "add x20, %x[args_ptr], %[offset_min]\n"
        "ld1r { v17.4s }, [x21]\n"
        "ld1r { v16.4s }, [x20]\n"
        "fmin v20.4s, v20.4s, v17.4s\n"
        "fmin v21.4s, v21.4s, v17.4s\n"
        "fmin v22.4s, v22.4s, v17.4s\n"
        "fmin v23.4s, v23.4s, v17.4s\n"
        "fmin v24.4s, v24.4s, v17.4s\n"
        "fmin v25.4s, v25.4s, v17.4s\n"
        "fmin v26.4s, v26.4s, v17.4s\n"
        "fmin v27.4s, v27.4s, v17.4s\n"
        "fmin v28.4s, v28.4s, v17.4s\n"
        "fmin v29.4s, v29.4s, v17.4s\n"
        "fmin v30.4s, v30.4s, v17.4s\n"
        "fmin v31.4s, v31.4s, v17.4s\n"
        "fmax v20.4s, v20.4s, v16.4s\n"
        "fmax v21.4s, v21.4s, v16.4s\n"
        "fmax v22.4s, v22.4s, v16.4s\n"
        "fmax v23.4s, v23.4s, v16.4s\n"
        "fmax v24.4s, v24.4s, v16.4s\n"
        "fmax v25.4s, v25.4s, v16.4s\n"
        "fmax v26.4s, v26.4s, v16.4s\n"
        "fmax v27.4s, v27.4s, v16.4s\n"
        "fmax v28.4s, v28.4s, v16.4s\n"
        "fmax v29.4s, v29.4s, v16.4s\n"
        "fmax v30.4s, v30.4s, v16.4s\n"
        "fmax v31.4s, v31.4s, v16.4s\n"
        "144:"  // Height 6: No activation
        "cmp x11, #0x8\n"
        "bge 149f\n"
        "tbz x11, #2, 146f\n"
        "st1 { v20.4s }, [x9], #0x10\n"
        "st1 { v22.4s }, [x26], #0x10\n"
        "st1 { v24.4s }, [x25], #0x10\n"
        "st1 { v26.4s }, [x24], #0x10\n"
        "st1 { v28.4s }, [x23], #0x10\n"
        "st1 { v30.4s }, [x22], #0x10\n"
        "tbz x11, #1, 145f\n"
        "str d21, [x9], #0x8\n"
        "str d23, [x26], #0x8\n"
        "str d25, [x25], #0x8\n"
        "str d27, [x24], #0x8\n"
        "str d29, [x23], #0x8\n"
        "str d31, [x22], #0x8\n"
        "tbz x11, #0, 148f\n"
        "st1 { v21.s }[2], [x9]\n"
        "st1 { v23.s }[2], [x26]\n"
        "st1 { v25.s }[2], [x25]\n"
        "st1 { v27.s }[2], [x24]\n"
        "st1 { v29.s }[2], [x23]\n"
        "st1 { v31.s }[2], [x22]\n"
        "b 148f\n"
        "145:"  // Height 6: Partial direct writeback: partial_1_4
        "tbz x11, #0, 148f\n"
        "str s21, [x9, #0x0]\n"
        "str s23, [x26, #0x0]\n"
        "str s25, [x25, #0x0]\n"
        "str s27, [x24, #0x0]\n"
        "str s29, [x23, #0x0]\n"
        "str s31, [x22, #0x0]\n"
        "b 148f\n"
        "146:"  // Height 6: Partial direct writeback: partial_2_0
        "tbz x11, #1, 147f\n"
        "str d20, [x9], #0x8\n"
        "str d22, [x26], #0x8\n"
        "str d24, [x25], #0x8\n"
        "str d26, [x24], #0x8\n"
        "str d28, [x23], #0x8\n"
        "str d30, [x22], #0x8\n"
        "tbz x11, #0, 148f\n"
        "st1 { v20.s }[2], [x9]\n"
        "st1 { v22.s }[2], [x26]\n"
        "st1 { v24.s }[2], [x25]\n"
        "st1 { v26.s }[2], [x24]\n"
        "st1 { v28.s }[2], [x23]\n"
        "st1 { v30.s }[2], [x22]\n"
        "b 148f\n"
        "147:"  // Height 6: Partial direct writeback: partial_1_0
        "str s20, [x9, #0x0]\n"
        "str s22, [x26, #0x0]\n"
        "str s24, [x25, #0x0]\n"
        "str s26, [x24, #0x0]\n"
        "str s28, [x23, #0x0]\n"
        "str s30, [x22, #0x0]\n"
        "148:"  // Height 6: Partial direct writeback: Done
        "b 150f\n"
        "149:"  // Height 6: Full writeback
        "str q20, [x9, #0x0]\n"
        "str q21, [x9, #0x10]\n"
        "add x9, x9, #0x20\n"
        "str q22, [x26, #0x0]\n"
        "str q23, [x26, #0x10]\n"
        "str q24, [x25, #0x0]\n"
        "str q25, [x25, #0x10]\n"
        "str q26, [x24, #0x0]\n"
        "str q27, [x24, #0x10]\n"
        "str q28, [x23, #0x0]\n"
        "str q29, [x23, #0x10]\n"
        "str q30, [x22, #0x0]\n"
        "str q31, [x22, #0x10]\n"
        "150:"  // Height 6: Writeback done
        "subs x11, x11, #0x8\n"
        "bgt 127b\n"
        "subs %x[m], %x[m], #0x6\n"
        "beq 152f\n"
        "ldr x21, [%x[args_ptr], %[offsetof_input_offset]]\n"
        "tbz %x[flags], #3, 151f\n"
        "add x21, x21, #0x6\n"
        "str x21, [%x[args_ptr], %[offsetof_input_offset]]\n"
        "b 1b\n"
        "151:"  // Update direct input
        "mov x20, #0x18\n"
        "madd %x[input_ptr], x20, x21, %x[input_ptr]\n"
        "b 1b\n"
        "152:"  // Exit
        : [input_ptr] "+&r"(input_ptr), [m] "+&r"(m)
        : [args_ptr] "r"(&ka), [flags] "r"(flags), [offset_max] "I"(offsetof(KernelArgs, maxval)),
          [offset_min] "I"(offsetof(KernelArgs, minval)), [offsetof_B_ptr] "I"(offsetof(KernelArgs, B_ptr)),
          [offsetof_N] "I"(offsetof(KernelArgs, N)),
          [offsetof_input_initial_col] "I"(offsetof(KernelArgs, input_initial_col)),
          [offsetof_input_offset] "I"(offsetof(KernelArgs, input_offset)),
          [offsetof_num_strings] "I"(offsetof(KernelArgs, num_strings)),
          [offsetof_output_offset] "I"(offsetof(KernelArgs, output_offset)),
          [offsetof_output_ptr] "I"(offsetof(KernelArgs, output_ptr)),
          [offsetof_string_lengths] "I"(offsetof(KernelArgs, string_lengths))
        : "cc", "memory", "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9", "v10", "v11", "v12", "v13", "v16",
          "v17", "v18", "v19", "v20", "v21", "v22", "v23", "v24", "v25", "v26", "v27", "v28", "v29", "v30", "v31", "x9",
          "x10", "x11", "x20", "x21", "x22", "x23", "x24", "x25", "x26", "x27", "x28");
}

#endif  // Architectural features check.
